	SHAbuf		PROTO	:DWORD,:DWORD,:DWORD
	Pad		PROTO	:DWORD
	SHA256		PROTO	:DWORD,:DWORD,:DWORD
;========================================================================================================================
.DATA
	; constants table
	K	dd	0428A2F98h,071374491h,0B5C0FBCFh,0E9B5DBA5h,03956C25Bh,059F111F1h,0923F82A4h,0AB1C5ED5h
		dd	0D807AA98h,012835B01h,0243185BEh,0550C7DC3h,072BE5D74h,080DEB1FEh,09BDC06A7h,0C19BF174h
		dd	0E49B69C1h,0EFBE4786h,00FC19DC6h,0240CA1CCh,02DE92C6Fh,04A7484AAh,05CB0A9DCh,076F988DAh
		dd	0983E5152h,0A831C66Dh,0B00327C8h,0BF597FC7h,0C6E00BF3h,0D5A79147h,006CA6351h,014292967h
		dd	027B70A85h,02E1B2138h,04D2C6DFCh,053380D13h,0650A7354h,0766A0ABBh,081C2C92Eh,092722C85h
		dd	0A2BFE8A1h,0A81A664Bh,0C24B8B70h,0C76C51A3h,0D192E819h,0D6990624h,0F40E3585h,0106AA070h
		dd	019A4C116h,01E376C08h,02748774Ch,034B0BCB5h,0391C0CB3h,04ED8AA4Ah,05B9CCA4Fh,0682E6FF3h
		dd	0748F82EEh,078A5636Fh,084C87814h,08CC70208h,090BEFFFAh,0A4506CEBh,0BEF9A3F7h,0C67178F2h
.DATA?
	; hash values
	H0	dd	?
	H1	dd	?
	H2	dd	?
	H3	dd	?
	H4	dd	?
	H5	dd	?
	H6	dd	?
	H7	dd	?

	; working variables
	a	dd	?
	b	dd	?
	c_	dd	?
	d	dd	?
	e	dd	?
	f	dd	?
	g	dd	?
	h	dd	?

	; message schedule
	W	dd	64 dup (?)

	; message blocks counters
	all_blocks	dd	?
	block_counter	dd	?
	pass_counter	dd	?

	pad_buffer	dd	?	; ptr to padded piece of message
	pad_lngth	dd	?	; length of this piece

.CODE
;========================================================================================================================
SHAbuf PROC pBuf:DWORD,buf_size:DWORD,pCallback:DWORD

	LOCAL	second_call_blocks	:DWORD
	LOCAL	pad_buf			:DWORD
	LOCAL	first_call_lngth	:DWORD

	invoke	Pad,	buf_size
	mov	eax,pad_lngth

	.IF	eax>64
		mov	second_call_blocks,2
		mov	ebx,128
	.ELSE
		mov	second_call_blocks,1
		mov	ebx,64
	.ENDIF

	sub	ebx,eax
	mov	ecx,buf_size
	sub	ecx,ebx
	mov	first_call_lngth,ecx
	shr	ecx,6

	mov	eax,ecx
	add	eax,second_call_blocks
	mov	all_blocks,eax		; progress maximum


	mrm	H0,06A09E667h
	mrm	H1,0BB67AE85h
	mrm	H2,03C6Ef372h
	mrm	H3,0A54FF53Ah
	mrm	H4,0510E527Fh
	mrm	H5,09B05688Ch
	mrm	H6,01F83D9ABh
	mrm	H7,05BE0CD19h


	invoke	SHA256,		pBuf,ecx,pCallback

	ALLOC	128
	mov	pad_buf,eax
	mov	edi,eax

	mov	ecx,64
	cmp	second_call_blocks,1
	je	@F
	add	ecx,64
@@:	sub	ecx,pad_lngth

	mov	esi,pBuf
	add	esi,first_call_lngth
	cmp	ecx,0
	je	@F
	rep	movsb

@@:	mov	ecx,pad_lngth
	mov	esi,pad_buffer
	rep	movsb

	invoke	SHA256,		pad_buf,second_call_blocks,pCallback


	mov	edi,pad_buf	; bigger buffer than needed, but it's allocated
	mov	esi,offset H0
	mov	ecx,8
@@:	lodsd
	bswap	eax
	stosd
	dec	ecx
	jnz	@B

	FREE	pad_buffer


	mov	eax,pad_buf			; 128 instead of 32 B, free this outside!!!
	ret

SHAbuf ENDP
;========================================================================================================================
Pad PROC	lngth:DWORD	; lngth - length of the whole message before padding, e.g. a file's size

	ALLOC	128
	mov	pad_buffer,eax
	mov	edi,eax

	mov	eax,lngth
	mov	ebx,eax
	shr	eax,6		; lngth/64, eax == full 512 bit blocks
	push	eax
	shl	eax,6
	sub	ebx,eax		; ebx == bytes over multiple of 512 bits [64 bytes]
	pop	eax
	inc	eax

	.IF	ebx>55		; 64 byte blocks - 9 obligatory
		mov	ecx,128
	.ELSE
		mov	ecx,64
	.ENDIF
	sub	ecx,ebx
	sub	ecx,9

	mov	BYTE PTR [edi],128	; 10000000
	inc	edi
	xor	al,al
	rep	stosb			; ecx zero bytes
	mov	eax,lngth
	cmp	eax,1FFFFFFFh		; max bytes that will fit a dword as bits
	ja	@F

	shl	eax,3
	mov	DWORD PTR [edi],0
	add	edi,4
	bswap	eax
	stosd
	sub	edi,pad_buffer
	mov	pad_lngth,edi
	ret
@@:
	mov	ebx,eax
	rol	ebx,3		; make last 3 digits first 3
	and	ebx,7		; 00000111
	shl	eax,3
	bswap	ebx
	mov	[edi],ebx
	add	edi,4
	bswap	eax
	stosd
	sub	edi,pad_buffer
	mov	pad_lngth,edi
	ret

Pad	ENDP
;========================================================================================================================
SHA256	PROC	pBuf:DWORD,blocks:DWORD,pCallback:DWORD

	;---------------------------------------------------------------------------------------
	;	pBuf		-	ptr to buffer containing Message
	;	blocks		-	full 64 byte blocks to be processed
	;	pCallback	-	Callback PROC blocks_processed:DWORD,all_blocks:DWORD,
	;				for progress indication, can be NULL
	;---------------------------------------------------------------------------------------

	cmp	blocks,0
	je	the_end
	mrm	block_counter,blocks
	mov	pass_counter,0



PROCESS_BLOCK:

	; prepare the message schedule
	mov	esi,pBuf
	mov	eax,pass_counter
	shl	eax,6
	add	esi,eax
	mov	edi,offset W
	mov	ecx,16
	rep	movsd		; W(t) = M(t),							t = <0,15>

	mov	ecx,48		; W(t) = Sigma1(W(t-2)) + W(t-7) + Sigma0(W(t-15)) + W(t-16)	t = <16,63>
@@:	mov	eax,[edi-60]
	bswap	eax

;-----------------------------------	invoke	Sigma0,	eax	== (ror x,7) XOR (ror x,18) XOR (shr x,3)
	mov	ebx,eax
	mov	edx,eax
	ror	eax,7
	rol	ebx,14
	shr	edx,3
	xor	eax,ebx
	xor	eax,edx
;-----------------------------------

	mov	ebx,[edi-64]
	bswap	ebx
	add	eax,ebx
	mov	ebx,[edi-28]
	bswap	ebx
	add	eax,ebx
	mov	edx,[edi-8]
	bswap	edx
	mov	esi,eax

;-----------------------------------	invoke	Sigma1,	edx	== (ror x,17) XOR (ror x,19) XOR (shr x,10)
	mov	eax,edx
	mov	ebx,edx
	rol	eax,15
	rol	ebx,13
	shr	edx,10
	xor	eax,ebx
	xor	eax,edx
;-----------------------------------

	add	esi,eax
	bswap	esi
	mov	[edi],esi
	add	edi,4
	dec	ecx
	jnz	@B



	; initialize working variables
	mrm	a,H0
	mrm	b,H1
	mrm	c_,H2
	mrm	d,H3
	mrm	e,H4
	mrm	f,H5
	mrm	g,H6
	mrm	h,H7



	sub	ecx,ecx
@@:
	; 64 *

	mov	esi,h			; T1 = h + Sum1(e) + Ch(e,f,g) + K(t) + W(t)

;-----------------------------------	invoke	Sum1,	e	== (ror x,6) XOR (ror x,11) XOR (ror x,25)
	mov	eax,e
	mov	ebx,eax
	mov	edx,eax
	ror	eax,6
	ror	ebx,11
	rol	edx,7
	xor	eax,ebx
	xor	eax,edx
;-----------------------------------

	add	esi,eax

;-----------------------------------	invoke	Ch_,	e,f,g	== (x AND y) XOR ((NOT x) AND z)
	mov	eax,e
	mov	ebx,eax
	not	ebx
	and	ebx,g
	and	eax,f
	xor	eax,ebx
;-----------------------------------

	add	esi,eax
	mov	ebx,ecx
	shl	ebx,2
	mov	eax,[ebx+K]
	add	esi,eax
	mov	eax,[ebx+W]
	bswap	eax
	add	esi,eax

;-----------------------------------	invoke	Sum0,	a	== (ror x,2) XOR (ror x,13) XOR (ror x,22)
	mov	eax,a
	mov	ebx,eax
	mov	edx,eax
	ror	eax,2
	ror	ebx,13
	rol	edx,10
	xor	eax,ebx
	xor	eax,edx
;-----------------------------------

	mov	edi,eax

;-----------------------------------	invoke	Maj,	a,b,c_	== (x AND y) XOR (x AND z) XOR (y AND z)
	mov	eax,a
	mov	edx,eax
	mov	ebx,b
	and	eax,ebx
	and	edx,c_
	and	ebx,c_
	xor	eax,ebx
	xor	eax,edx
;-----------------------------------

	add	edi,eax			; T2 = Sum0(a) + Maj(a,b,c)

	mrm	h,g
	mrm	g,f
	mrm	f,e

	mov	eax,d		; e = d + T1
	add	eax,esi
	mov	e,eax

	mrm	d,c_
	mrm	c_,b
	mrm	b,a

	add	esi,edi		; a = T1 + T2
	mov	a,esi

	inc	ecx
	cmp	ecx,64
	jne	@B


	; compute the intermediate/final hash value
	mov	eax,a		; H0(t) = a + H0(t-1)
	add	H0,eax
	mov	eax,b		; H1(t) = b + H1(t-1)
	add	H1,eax
	mov	eax,c_		; H2(t) = c + H2(t-1)
	add	H2,eax
	mov	eax,d		; H3(t) = d + H3(t-1)
	add	H3,eax
	mov	eax,e		; H4(t) = e + H4(t-1)
	add	H4,eax
	mov	eax,f		; H5(t) = f + H5(t-1)
	add	H5,eax
	mov	eax,g		; H6(t) = g + H6(t-1)
	add	H6,eax
	mov	eax,h		; H7(t) = h + H7(t-1)
	add	H7,eax

	inc	pass_counter
	cmp	pCallback,0
	je	@F

		push	all_blocks
		push	pass_counter
		CALL	pCallback

@@:	dec	block_counter
	jnz	PROCESS_BLOCK

the_end:	ret

SHA256	ENDP
;========================================================================================================================
END